/*
 * Copyright 2012 htfv (Aliaksei Lahachou)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.github.htfv.maven.plugins.testing.internal;

import java.util.Enumeration;
import java.util.NoSuchElementException;

/**
 * Decorates another {@link Enumeration} to filter elements.
 *
 * @param <E>
 *            the type of {@link Enumeration} elements.
 *
 * @author htfv (Aliaksei Lahachou)
 */
public abstract class AbstractFilteredEnumeration<E> implements Enumeration<E>
{
    /**
     * The filters {@link Enumeration}.
     */
    private Enumeration<E> enumeration;

    /**
     * The next object in the {@link Enumeration}.
     */
    private E nextElement;

    /**
     * The flag indicating whether the next element was initialized.
     */
    private boolean nextElementInitialized;

    /**
     * Constructs a new {@link AbstractFilteredEnumeration} object.
     *
     * @param enumeration
     *            the {@link Enumeration} to filter.
     */
    public AbstractFilteredEnumeration(final Enumeration<E> enumeration)
    {
        this.enumeration = enumeration;
    }

    @Override
    public boolean hasMoreElements()
    {
        if (nextElementInitialized)
        {
            return true;
        }

        return initializeNextElement();
    }

    /**
     * Finds the next element from the {@link Enumeration}.
     *
     * @return {@code true} if the next element exists; {@code false} otherwise.
     */
    private boolean initializeNextElement()
    {
        while (enumeration.hasMoreElements())
        {
            E localNextElement = enumeration.nextElement();

            if (shouldIncludeElement(localNextElement))
            {
                nextElement            = localNextElement;
                nextElementInitialized = true;

                return true;
            }
        }

        return false;
    }

    @Override
    public E nextElement()
    {
        if (!nextElementInitialized && !initializeNextElement())
        {
            throw new NoSuchElementException();
        }

        nextElementInitialized = false;

        return nextElement;
    }

    /**
     * Checks whether the specified element should be included in the result
     * {@link Enumeration}.
     *
     * @param element
     *            the element to check.
     *
     * @return {@code true} if the element should be included; {@code false}
     *         otherwise.
     */
    protected abstract boolean shouldIncludeElement(E element);
}
